@@ -43,6 +43,7 @@
 
 #define POLL_CLCC_INTERVAL 2000
 #define POLL_CLCC_DELAY 50
+#define EXPECT_RELEASE_DELAY 50
 #define CLIP_TIMEOUT 500
 
 static const char *none_prefix[] = { NULL };
@@ -57,6 +58,7 @@ struct voicecall_data {
 	int cind_val[HFP_INDICATOR_LAST];
 	unsigned int local_release;
 	unsigned int clcc_source;
+	unsigned int expect_release_source;
 	unsigned int clip_source;
 };
 
@@ -91,6 +93,14 @@ static GSList *find_dialing(GSList *calls)
 	return c;
 }
 
+static void voicecall_notify(gpointer value, gpointer user)
+{
+	struct ofono_call *call = value;
+	struct ofono_voicecall *vc = user;
+
+	ofono_voicecall_notify(vc, call);
+}
+
 static struct ofono_call *create_call(struct ofono_voicecall *vc, int type,
 					int direction, int status,
 					const char *num, int num_type, int clip)
@@ -123,19 +133,6 @@ static struct ofono_call *create_call(struct ofono_voicecall *vc, int type,
 	return call;
 }
 
-static struct ofono_call *new_call_notify(struct ofono_voicecall *vc, int type,
-					int direction, int status,
-					const char *num, int num_type, int clip)
-{
-	struct ofono_call *c;
-
-	c = create_call(vc, type, direction, status, num, num_type, clip);
-
-	ofono_voicecall_notify(vc, c);
-
-	return c;
-}
-
 static void release_call(struct ofono_voicecall *vc, struct ofono_call *call)
 {
 	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
@@ -199,6 +196,11 @@ static void release_with_status(struct ofono_voicecall *vc, int status)
 		c = c->next;
 		g_slist_free_1(t);
 	}
+
+	if (vd->expect_release_source) {
+		g_source_remove(vd->expect_release_source);
+		vd->expect_release_source = 0;
+	}
 }
 
 static void clcc_poll_cb(gboolean ok, GAtResult *result, gpointer user_data)
@@ -210,11 +212,13 @@ static void clcc_poll_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	struct ofono_call *nc, *oc;
 	unsigned int num_active = 0;
 	unsigned int num_held = 0;
+	GSList *notify_calls = NULL;
+	unsigned int mpty_ids;
 
 	if (!ok)
 		return;
 
-	calls = at_util_parse_clcc(result);
+	calls = at_util_parse_clcc(result, &mpty_ids);
 
 	n = calls;
 	o = vd->calls;
@@ -247,7 +251,7 @@ static void clcc_poll_cb(gboolean ok, GAtResult *result, gpointer user_data)
 		} else if (nc && (oc == NULL || (nc->id < oc->id))) {
 			/* new call, signal it */
 			if (nc->type == 0)
-				ofono_voicecall_notify(vc, nc);
+				notify_calls = g_slist_append(notify_calls, nc);
 
 			n = n->next;
 		} else {
@@ -262,13 +266,24 @@ static void clcc_poll_cb(gboolean ok, GAtResult *result, gpointer user_data)
 
 			if (memcmp(nc, oc, sizeof(struct ofono_call)) &&
 					!nc->type)
-				ofono_voicecall_notify(vc, nc);
+				notify_calls = g_slist_prepend(notify_calls,
+									nc);
 
 			n = n->next;
 			o = o->next;
 		}
 	}
 
+	/*
+	 * Disconnections were already reported, so process the rest of the
+	 * notifications. Note that the new calls are placed at the end of the
+	 * list, after other state changes
+	 */
+	g_slist_foreach(notify_calls, voicecall_notify, vc);
+	g_slist_free(notify_calls);
+
+	ofono_voicecall_mpty_hint(vc, mpty_ids);
+
 	g_slist_foreach(vd->calls, (GFunc) g_free, NULL);
 	g_slist_free(vd->calls);
 
@@ -418,8 +433,13 @@ static void hfp_answer(struct ofono_voicecall *vc,
 static void hfp_hangup(struct ofono_voicecall *vc,
 			ofono_voicecall_cb_t cb, void *data)
 {
+	unsigned int affected = (1 << CALL_STATUS_INCOMING) |
+				(1 << CALL_STATUS_DIALING) |
+				(1 << CALL_STATUS_ALERTING) |
+				(1 << CALL_STATUS_ACTIVE);
+
 	/* Hangup current active call */
-	hfp_template("AT+CHUP", vc, generic_cb, 0x1, cb, data);
+	hfp_template("AT+CHUP", vc, generic_cb, affected, cb, data);
 }
 
 static void hfp_hold_all_active(struct ofono_voicecall *vc,
@@ -454,8 +474,7 @@ static void hfp_set_udub(struct ofono_voicecall *vc,
 			ofono_voicecall_cb_t cb, void *data)
 {
 	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
-	unsigned int incoming_or_waiting =
-		(1 << CALL_STATUS_INCOMING) | (1 << CALL_STATUS_WAITING);
+	unsigned int incoming_or_waiting = 1 << CALL_STATUS_WAITING;
 
 	if (vd->ag_mpty_features & AG_CHLD_0) {
 		hfp_template("AT+CHLD=0", vc, generic_cb, incoming_or_waiting,
@@ -466,13 +485,52 @@ static void hfp_set_udub(struct ofono_voicecall *vc,
 	CALLBACK_WITH_FAILURE(cb, data);
 }
 
+static gboolean expect_release(gpointer user_data)
+{
+	struct ofono_voicecall *vc = user_data;
+	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
+
+	g_at_chat_send(vd->chat, "AT+CLCC", clcc_prefix,
+				clcc_poll_cb, vc, NULL);
+
+	vd->expect_release_source = 0;
+
+	return FALSE;
+}
+
+static void release_all_active_cb(gboolean ok, GAtResult *result,
+							gpointer user_data)
+{
+	struct change_state_req *req = user_data;
+	struct voicecall_data *vd = ofono_voicecall_get_data(req->vc);
+
+	if (!ok)
+		goto out;
+
+	if (vd->expect_release_source)
+		g_source_remove(vd->expect_release_source);
+
+	/*
+	 * Some phones, like Nokia 500, do not send CIEV after accepting
+	 * the CHLD=1 command, even though the spec states that they should.
+	 * So simply poll to force the status update if the AG is misbehaving.
+	 */
+	vd->expect_release_source = g_timeout_add(EXPECT_RELEASE_DELAY,
+							expect_release,
+							req->vc);
+
+out:
+	generic_cb(ok, result, user_data);
+}
+
 static void hfp_release_all_active(struct ofono_voicecall *vc,
 					ofono_voicecall_cb_t cb, void *data)
 {
 	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
 
 	if (vd->ag_mpty_features & AG_CHLD_1) {
-		hfp_template("AT+CHLD=1", vc, generic_cb, 0x1, cb, data);
+		hfp_template("AT+CHLD=1", vc, release_all_active_cb, 0x1, cb,
+									data);
 		return;
 	}
 
@@ -849,59 +907,6 @@ static void ciev_call_notify(struct ofono_voicecall *vc,
 	vd->cind_val[HFP_INDICATOR_CALL] = value;
 }
 
-static void sync_dialing_cb(gboolean ok, GAtResult *result, gpointer user_data)
-{
-	struct ofono_voicecall *vc = user_data;
-	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
-	struct ofono_error error;
-	GSList *calls;
-	GSList *o;
-	GSList *n;
-	struct ofono_call *oc;
-	struct ofono_call *nc;
-
-	decode_at_error(&error, g_at_result_final_response(result));
-
-	if (!ok)
-		return;
-
-	calls = at_util_parse_clcc(result);
-
-	if (calls == NULL)
-		return;
-
-	/* Look for dialing or alerting calls on the new list */
-	n = find_dialing(calls);
-
-	/* Let us find if we have done the dial from HF by looking for
-	 * existing dialing or alerting calls
-	 */
-	o = find_dialing(vd->calls);
-
-	if (n == NULL && o) {
-		oc = o->data;
-		release_call(vc, oc);
-		vd->calls = g_slist_remove(vd->calls, oc);
-	} else if (n && o == NULL) {
-		nc = n->data;
-		new_call_notify(vc, nc->type, nc->direction, nc->status,
-				nc->phone_number.number, nc->phone_number.type,
-				nc->clip_validity);
-	} else if (n && o) {
-		oc = o->data;
-		nc = n->data;
-
-		memcpy(&oc->phone_number, &nc->phone_number,
-				sizeof(struct ofono_phone_number));
-		oc->status = nc->status;
-		oc->clip_validity = nc->clip_validity;
-		ofono_voicecall_notify(vc, oc);
-	}
-
-	g_slist_foreach(calls, (GFunc) g_free, NULL);
-	g_slist_free(calls);
-}
-
 static void ciev_callsetup_notify(struct ofono_voicecall *vc,
 					unsigned int value)
 {
@@ -979,7 +984,7 @@ static void ciev_callsetup_notify(struct ofono_voicecall *vc,
 		 * from AG: query and create call.
 		 */
 		g_at_chat_send(vd->chat, "AT+CLCC", clcc_prefix,
-				sync_dialing_cb, vc, NULL);
+				clcc_poll_cb, vc, NULL);
 		break;
 
 	case 3:
@@ -1095,15 +1100,15 @@ static void hfp_clcc_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	struct ofono_voicecall *vc = user_data;
 	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
-	GSList *l;
+	unsigned int mpty_ids;
 
 	if (!ok)
 		return;
 
-	vd->calls = at_util_parse_clcc(result);
+	vd->calls = at_util_parse_clcc(result, &mpty_ids);
 
-	for (l = vd->calls; l; l = l->next)
-		ofono_voicecall_notify(vc, l->data);
+	g_slist_foreach(vd->calls, voicecall_notify, vc);
+	ofono_voicecall_mpty_hint(vc, mpty_ids);
 }
 
 static void hfp_voicecall_initialized(gboolean ok, GAtResult *result,
@@ -1161,6 +1166,9 @@ static void hfp_voicecall_remove(struct ofono_voicecall *vc)
 	if (vd->clip_source)
 		g_source_remove(vd->clip_source);
 
+	if (vd->expect_release_source)
+		g_source_remove(vd->expect_release_source);
+
 	g_slist_foreach(vd->calls, (GFunc) g_free, NULL);
 	g_slist_free(vd->calls);
 
